package com.exh.bevel;

import java.io.File;
import java.io.FileOutputStream;
import java.util.LinkedList;
import java.util.List;

import android.content.Context;
import android.content.res.Resources;
import android.graphics.Bitmap;
import android.graphics.BitmapFactory;
import android.graphics.Canvas;
import android.graphics.Paint;
import android.graphics.Point;
import android.os.Environment;

import com.exh.bevel.render.RenderTri;
import com.exh.math.primitive.Point3;

/* DOC-GENERATOR.
 * Just a class to handle generating bitmaps to include in the documentation.
 */
public final class DocGenerator implements Generator {

	private final List<Bitmap>		mList = new LinkedList<Bitmap>();
	private final Bitmap			mArrow;
	private final ShapeGenerator	mSilo = new ShapeGenerator();
	
	private final Point				mSize = new Point();
	
	private final static String		DIR = "bevel";
	private final int				PAD = 20;
	
	public DocGenerator(final Context ctx) {
		final Resources		res = ctx.getApplicationContext().getResources();
		mArrow = BitmapFactory.decodeResource(res, R.drawable.arrow);
	}
	
	public Bitmap run(final ProgressReport p) {
		p.setProgressReport(0);

		final float					SCALE = 0.5f;
		final float					STACK_SCALE = 0.25f;		
		save("00.png", newEvolution(p), STACK_SCALE);
		// Source polygon
		save("01.png", newPoly(Input.initSilo()), SCALE);
		// Circle, small bevel, stroked
		save("02.png", mSilo.run(Input.initCircle(20, 0), 30, RenderTri.STROKE, p), SCALE);		
		// Silo, small bevel, stroked
		save("03.png", mSilo.run(Input.initSilo(), 30, RenderTri.STROKE, p), SCALE);
		// Silo, large bevel, no clipping, stroked
		save("04.png", mSilo.run(Input.initSilo(), 90, RenderTri.STROKE | RenderTri.NO_CLIP, p), SCALE);		
		// Silo, large bevel, stroked
		save("05.png", mSilo.run(Input.initSilo(), 90, RenderTri.STROKE, p), SCALE);		
		// Silo, large bevel, filled
		save("06.png", mSilo.run(Input.initSilo(), 90, RenderTri.FILL | RenderTri.STROKE, p), SCALE);		
		// Silo, large bevel, rendered
		save("07.png", mSilo.run(Input.initSilo(), 90, RenderTri.FILL | RenderTri.DEPTH, p), SCALE);		
		// Star, large bevel, rendered
		save("08.png", mSilo.run(Input.initThrowingStar(), 90, RenderTri.FILL | RenderTri.DEPTH, p), SCALE);		
		// T, large bevel, rendered
		save("09.png", mSilo.run(Input.initT(), 90, RenderTri.FILL | RenderTri.DEPTH, p), SCALE);		
		// Tree, large bevel, rendered
		save("10.png", mSilo.run(Input.initTree(), 90, RenderTri.FILL | RenderTri.DEPTH, p), SCALE);		
		// Low res sphere
		save("28.png", newSphereEvolution(10, 120, p), STACK_SCALE);
		// Hi res sphere
		return save("29.png", newSphereEvolution(30, 120, p), STACK_SCALE);
	}

	// The evolution
	private Bitmap newEvolution(final ProgressReport p) {
		mList.clear();
		
		mList.add(newPoly(Input.initSilo()));
		mList.add(mArrow);
		mList.add(mSilo.run(Input.initSilo(), 30, RenderTri.STROKE, p));
		mList.add(mArrow);
		mList.add(mSilo.run(Input.initSilo(), 90, RenderTri.STROKE, p));
		mList.add(mArrow);
		mList.add(mSilo.run(Input.initSilo(), 90, RenderTri.FILL | RenderTri.STROKE, p));

		return assembleHrz(PAD);
	}

	// Low res sphere
	private Bitmap newSphereEvolution(final int spokes, final double amt, final ProgressReport p) {
		mList.clear();
		
		mList.add(mSilo.run(Input.initCircle(spokes, 0), amt, RenderTri.STROKE, p));
		mList.add(mArrow);
		mList.add(mSilo.run(Input.initCircle(spokes, 0), amt, RenderTri.FILL | RenderTri.STROKE | RenderTri.DEPTH, p));

		return assembleHrz(PAD);
	}

	// Combine all my bitmaps into one, layout horizontal
	public Bitmap assembleHrz(final int pad) {
		if (mList.size() < 1) return null;
		if (mList.size() == 1) return mList.get(0);

		int					w = 0, h = 0;
		for (int k=0; k<mList.size(); k++) {
			final Bitmap	bm = mList.get(k);
			if (k > 0) w += pad;
			w += bm.getWidth();
			if (bm.getHeight() > h) h = bm.getHeight();
		}
		final Bitmap		out = Bitmap.createBitmap(w, h, Bitmap.Config.ARGB_8888);
		out.eraseColor(0x00000000);
		final Canvas		canvas = new Canvas(out);
		
		int					x = 0;
		for (int k=0; k<mList.size(); k++) {
			final Bitmap	bm = mList.get(k);
			int				y = 0;
			if (bm.getHeight() < h) y = (h - bm.getHeight()) / 2;
			canvas.drawBitmap(bm, x, y, null);
			x += bm.getWidth() + pad;
		}
		return out;
	}

	private Bitmap newPoly(final Input pts) {
		ShapeGenerator.translateToBounds(pts.mPts, mSize);
		final Bitmap		bm = Bitmap.createBitmap(mSize.x, mSize.y, Bitmap.Config.ARGB_8888);
		bm.eraseColor(0x00000000);
		final Canvas		canvas = new Canvas(bm);
		
		final Paint			p = new Paint(Paint.ANTI_ALIAS_FLAG);
		
		p.setStrokeWidth(RenderTri.STROKE_THICK);
		p.setColor(RenderTri.FG_C);
		Point3				prv = pts.mPts.get(pts.mPts.size()-1);
		for (Point3 pt : pts.mPts) {
			canvas.drawLine((float)prv.x, (float)prv.y, (float)pt.x, (float)pt.y, p);
			prv = pt;
		}
		
		p.setStrokeWidth(1);
		p.setColor(0xff333333);
		for (Point3 pt : pts.mPts) {
			canvas.drawCircle((float)pt.x, (float)pt.y, 4, p);
		}

		return bm;
	}

	// FILE IO
	// -------

	// Save the bitmap to the SD card
	public static final Bitmap save(String filename, final Bitmap inBm, final float scale) {
		if (!verify(DIR)) return null;
		filename = DIR + File.separator + filename;
		if (inBm == null || filename.length() < 1) return null;
		
		Bitmap				bm = inBm;
		// Eh?  Doesn't resample the image??
/*
		if (scale < 1) {
			bm = Bitmap.createScaledBitmap(inBm, (int)(bm.getWidth()*scale), (int)(bm.getHeight()*scale), true);
			inBm.recycle();
		}
*/
		
		FileOutputStream	file = null;
		try {
			File			root = android.os.Environment.getExternalStorageDirectory();
			file = new FileOutputStream(root.getPath() + File.separator + filename);
			bm.compress(Bitmap.CompressFormat.PNG, 0, file);
			return bm;
		} catch (Exception ex) {
			System.out.println("ERROR saving=" + ex.toString());
		} finally {
			try { file.close(); } catch (Exception ex) { }
		}
		return null;
	}
	
	// Create the directory if necessary
	public static final boolean verify(final String dir) {
		if (!Environment.MEDIA_MOUNTED.equals(Environment.getExternalStorageState())) {
			System.out.println("ERROR no SD card mounted");
			return false;
		}
		try {
			final File		root = android.os.Environment.getExternalStorageDirectory();
			final File		f = new File(root, dir);
			if (!f.exists()) return f.mkdir();
			return f.canWrite();
		} catch (Exception ex) {
			System.out.println("ERROR verifying dir " + dir + " = " + ex);
		}
		return false;
	}

}
