package com.exh.math.poly;

import com.exh.math.primitive.Point3;
import com.exh.math.primitive.Polygon;
import com.exh.math.tests.LineLine;
import com.exh.math.tests.PtToSeg3;

// A helper class for Bool that takes a dst poly, and
// a src poly completely inside the dst, and returns a new
// poly that is a single path including both the dst and
// src, i.e. the src cut out of the dst.
class BoolCutOut {
	
	private final Hit[]		mHit = newHit(2);
	
	private final PtToSeg3	mPtS = new PtToSeg3();
	private final LineLine	mLL = new LineLine();
	
	boolean on(	final Polygon dst, final Polygon src,
				final Polygon out0, final Polygon out1) {
		out0.clear();
		out1.clear();
		// Find the starting and intersection points on each poly
		if (!getStarts(dst, src)) return false;
		// I've now split the combined poly into two, so add those
		// to the output.
		addPoly0(dst, src, out0);
		addPoly1(dst, src, out1);
		return true;
	}

	// Add half the final poly
	private void addPoly0(	final Polygon dst, final Polygon src,
							final Polygon out) {
		out.add(mHit[0].mPt);
		addPoly(dst, mHit[0].mDst, mHit[1].mDst, 1, out, true);
		out.add(mHit[1].mPt);
		addPoly(src, mHit[1].mSrc, mHit[0].mSrc, -1, out, true);
	}

	// Add the other half the final poly
	private void addPoly1(	final Polygon dst, final Polygon src,
							final Polygon out) {
		out.add(mHit[0].mPt);
		addPoly(dst, clip(mHit[0].mDst-1, dst), clip(mHit[1].mDst+0, dst), -1, out, false);
		out.add(mHit[1].mPt);
		addPoly(src, mHit[1].mSrc, mHit[0].mSrc, 1, out, true);
	}

	private void addPoly(	final Polygon p, final int start, final int end,
							final int delta, final Polygon out, final boolean close) {
		int				idx = start;
		do {
			out.add(p.mPt[idx]);
			idx = clip(idx+delta, p);
		} while (idx != end);
		if (close && start != end) out.add(p.mPt[end]);
	}

	private final static int clip(final int idx, final Polygon p) {
		if (idx >= p.mSize) return 0;
		if (idx < 0) return p.mSize-1;
		return idx;
	}
	
	private boolean getStarts(final Polygon dst, final Polygon src) {
		if (src.mSize < 2) return false;
		for (int k=0; k<mHit.length; k++) mHit[k].clear();
	
		// Take one side of the src poly and find its intersections
		// with the dst.  This is not great -- any complex src polys
		// will self intersect and be all messed up, but it works for
		// what I'm using it for.
		final Point3		a0 = src.mPt[0], a1 = src.mPt[1];
		Point3				prvB = dst.mPt[dst.mSize-1];
		for (int k=0; k<dst.mSize; k++) {
			final Point3	nxtB = dst.mPt[k];
			double			d = mLL.on(a0, a1, prvB, nxtB);
			// Make sure the intersection point is on the line and
			// also on the actual segment.
			if (!Double.isNaN(d) && d < 0.1 && mPtS.on(mLL.mNearest2, prvB, nxtB, null) < 0.1) {
				// Get distance from intersection to src segment, and
				// replace the lowest stored value.
				d = mPtS.on(mLL.mNearest2, a0, a1, null);
				final int	idx = highestD();
				if (idx >= 0 && (mHit[idx].mD < 0 || d < mHit[idx].mD)) {
					mHit[idx].mD = d;
					mHit[idx].mPt.set(mLL.mNearest2);
					mHit[idx].mDst = k;
				}
			}
		}
		for (int k=0; k<mHit.length; k++) {
			if (mHit[k].mDst < 0) return false;
		}
		mHit[0].mSrc = 0;
		mHit[1].mSrc = 1;
		// I now have two intersection (hit) points, so select the
		// one closest to the relevent src point.
		final double		a0_d = mHit[0].mPt.dist3Squared(a0),
							a1_d = mHit[0].mPt.dist3Squared(a1);
		// If intersection at hit[0] is closest to src pt 0 then 1,
		// everything's fine, otherwise switch the hits.
		if (a1_d < a0_d) {
			final Hit		h = mHit[0];	mHit[0] = mHit[1];	mHit[1] = h;
		}
		
		return true;
	}
	
	// Answer the mHit with the highest mD (-1 for off).
	private int highestD() {
		int			idx = -1;
		double		d = -1;
		for (int k=0; k<mHit.length; k++) {
			if (mHit[k].mD < 0) return k;
			if (idx < 0 || mHit[k].mD > d) {
				idx = k;
				d = mHit[k].mD;
			}
		}
		return idx;
	}

	// HIT class
	// ---------
	private final static class Hit {
		// Indexes into the dst and src polygon.
		int					mDst, mSrc;
		// The intersection on dst that the src points found
		// as the closest.
		final Point3		mPt = new Point3();
		
		// Tmp during setup
		double				mD;
		
		void clear() {
			mDst = -1;
			mSrc = -1;
			mD = -1;
		}
	}
	
	private final static Hit[]	newHit(final int size) {
		final Hit[]			ans = new Hit[size];
		for (int k=0; k<size; k++) ans[k] = new Hit();
		return ans;
	}

}
