package com.exh.math.poly;

import java.util.Comparator;
import java.util.TreeSet;

import com.exh.math.primitive.Point3;
import com.exh.math.primitive.Polygon;
import com.exh.math.tests.SegSeg;

/* The behaviour for building the flat step
 * lists from source polygons.
 */
class Flatten {

	// Eliminate duplicate points, based on a very fuzzy notion of
	// how different they can be.  This can greatly affect what gets
	// clipped, although honestly, as I've gotten the rest of the
	// algo tighter this seems like it works better with lower numbers,
	// as it should.
	double								mEps = 0.1;
	
	private final IntersectionList		mInter = new IntersectionList();
	// Source object for the intersections.  If it's a poly, then
	// the intersections are generated against the line segments.
	// If it's a StepList, I can just directly look up the existing
	// intersections.
	private Polygon						mPoly;
	private StepList					mList;

	// When there are multiple intersections across the same segment,
	// they need to be ordered correctly.
	private final PointComparator		mPtCmp = new PointComparator();
	private final TreeSet<Intersection>	mSort = new TreeSet<Intersection>(mPtCmp);

	private final SegSeg				mSS = new SegSeg();
	
	Flatten() {
		mSS.mEps = 0.001;
	}
	
	void start() {
		mInter.clear();
	}
	
	// Set to one or the other, based on whether you want intersections
	// generated or looked up.
	void setFindIntersectionSource(final Polygon p, final StepList l) {
		mPoly = p;
		mList = l;
	}

	void addIntersections(final Point3 a, final Point3 b, final StepList list) {
		mSort.clear();
		mPtCmp.setOrigin(a);
		
		if (mPoly != null) getNewIntersections(a, b);
		else if (mList != null) getExistingIntersections(a, b);
		
		for (Intersection i : mSort) list.add(i, mEps);
		mSort.clear();
	}
	
	// Find all intersections between the segment and mPoly.
	private void getNewIntersections(final Point3 a, final Point3 b) {
		Point3							prv = mPoly.mPt[mPoly.mSize-1];
		for (int k=0; k<mPoly.mSize; k++) {
			final Point3	nxt = mPoly.mPt[k];
//System.out.println("INTER BETWEEN between " + a + ", " + b + " and " + prv + ", " + nxt + " -pre=" + mSS.mEps);
			mSS.on(a, b, prv, nxt);
//System.out.println("\tdist=" + mSS.mDistanceToHit);
			if (mSS.on(a, b, prv, nxt)) {
//System.out.println("\thit " + mSS.mHit);
				final Intersection		inter = mInter.getNew();
				inter.set(mSS.mHit, a, b, prv, nxt);
				mSort.add(inter);
			}
			prv = nxt;
		}
	}
	
	// The intersections have already been created, so just find them.
	private void getExistingIntersections(final Point3 a, final Point3 b) {
		for (int k=0; k<mList.mSize; k++) {
			final Intersection		inter = mList.mStep[k].mIntersection;
			if (inter == null) continue;
			// The intersection was added with the source segment as a,
			// so I need to check the reverse.
			if ( (a == inter.mB[0] && b == inter.mB[1]) || (a == inter.mB[1] && b == inter.mB[0])) {
				mSort.add(inter);
			}
		}
	}

	// POINT-COMPARATOR
	// For comparing point distance from an origin
	// --------------------------------------------
	private final static class PointComparator implements Comparator<Intersection> {
		
		private Point3			mPt = new Point3();
		
		public void setOrigin(final Point3 pt) {
			mPt.set(pt);
		}
		
 		public int compare(Intersection o1, Intersection o2) {
 			final double	d1 = mPt.dist3Squared(o1.mPt),
 							d2 = mPt.dist3Squared(o2.mPt);
 			if (d1 < d2) return -1;
 			if (d1 == d2) return 0;
 			return 1;
 		}
 	}
}
