package com.exh.math.poly;

import com.exh.math.primitive.Point3;
import com.exh.math.primitive.Polygon;
import com.exh.math.primitive.Tri;

// Based on poly_tri, by Evans & Sutherland
// http://www.gamedev.net/reference/articles/article425.asp
public class Triangulizer {

	private Tri[]					mTri = new Tri[64];
	private int						mTriSize = 0;
	// Indexes into the vertex list
	private int[]					mVp = new int[64];

	private final static double		BIG = 1.0e30;
	
	public final static int			CCW = 0;
	public final static int			CW = 1;
	
	public Tri get(final int idx) {
		if (idx < 0 || idx >= mTriSize) return null;
		return mTri[idx];
	}
	
	public void clear() {
		if (mTri != null) {
			for (int k=0; k<mTri.length; k++) mTri[k] = null;
		}
		mTriSize = 0;
	}
	
	// Answer the number of triangles generated
	public int on(final Polygon poly) {
		mTriSize = 0;
		if (poly == null || poly.mSize < 3) return 0;
		// XXX My most common cases are a poly of size 3 or 4, both
		// trivial to triangulize, so I should hardcode.
		
		// Setup
		if (mTri.length < poly.mSize) mTri = new Tri[poly.mSize];
		if (mVp.length < poly.mSize) mVp = new int[poly.mSize];
		for (int k=0; k<poly.mSize; k++) mVp[k] = k;
		
		final int[]				vp = mVp;
		final Point3[]			v = poly.mPt;
	
		final int				poly_orientation = orientation(poly.mPt, poly.mSize);

		/* Slice off clean triangles until nothing remains */
		int						count = poly.mSize,
								next, prev,
								min_vert;
		double					dist, min_dist;
		while (count > 3) {
			min_dist = BIG;
			min_vert = 0;           /* Just in case we don't find one... */
			for (int cur = 0; cur < count; cur++) {
				prev = cur - 1;
				next = cur + 1;
				if (cur == 0) prev = count - 1;
				else if (cur == count - 1) next = 0;
				/* Pick out shortest distance that forms a good triangle */
				if (   (determinant( vp[prev], vp[cur], vp[next], v ) == poly_orientation)
						/* Same orientation as polygon */
						&& noInterior( vp[prev], vp[cur], vp[next], v, vp, count, poly_orientation )
						/* No points inside */
						&& ((dist = distSquared(v[vp[prev]].x, v[vp[prev]].y, v[vp[next]].x, v[vp[next]].y )) < min_dist) ) {
				/* Better than any so far */
					min_dist = dist;
					min_vert = cur;
				}
			} /* End of for each vertex (cur) */

			/* The following error should "never happen". */
			if (min_dist == BIG) {
				System.out.println("ERROR in Triangulizer, didn't find a triangle");
			}
			prev = min_vert - 1;
			next = min_vert + 1;
			if (min_vert == 0) prev = count - 1;
			else if (min_vert == count - 1) next = 0;

			// Found a triangle
			addTri(v[vp[prev]], v[vp[min_vert]], v[vp[next]]);

			/* Remove the triangle from the polygon */
			count -= 1;
			for (int i = min_vert; i < count; i++) {
				vp[i] = vp[i+1];
			}
		}

		// Final triangle
		addTri(v[vp[0]], v[vp[1]], v[vp[2]]);
		
		return mTriSize;
	}
	
	private final void addTri(final Point3 a, final Point3 b, final Point3 c) {
		mTri[mTriSize] = Tri.create(a, b, c);
		mTriSize++;
	}
	
	// Ignore that slow square root when we're only comparing
	// relative distances.
	public final static double distSquared(	final double x0, final double y0,
											final double x1, final double y1) {
		return ((x0-x1)*(x0-x1)) + ((y0-y1)*(y0-y1));
	}

	// Answer either CW or CCW for a polygon
	public final static int orientation(final Point3[] pts, final int size) {
		double		area;

		/* Do the wrap-around first */
		area = pts[size-1].x * pts[0].y - pts[0].x * pts[size-1].y;

		/* Compute the area (times 2) of the polygon */
		for (int i = 0; i < size-1; i++) {
			area += pts[i].x * pts[i+1].y - pts[i+1].x * pts[i].y;
		}
		if (area >= 0.0) return CCW;
		return CW;
	}

	public final static int signedOrientation(final Point3[] pts, final int size) {
		// Need to set origin at 0, because negative values screw up
		// the test.  NO!  WRONG!  NO DIFFERENCE!
		double		minX = pts[0].x, minY = pts[0].y;
		for (int k=1; k<size; k++) {
			if (pts[k].x < minX) minX = pts[k].x;
			if (pts[k].y < minY) minY = pts[k].y;
		}
		
		double		area;

		/* Do the wrap-around first */
		area = (pts[size-1].x-minX) * (pts[0].y-minY) - (pts[0].x-minX) * (pts[size-1].y-minY);

		/* Compute the area (times 2) of the polygon */
		for (int i = 0; i < size-1; i++) {
			area += (pts[i].x-minX) * (pts[i+1].y-minY) - (pts[i+1].x-minX) * (pts[i].y-minY);
		}
		if (area >= 0.0) return CCW;
		return CW;
	}

	// Compute the determinant of the three points.
	// Returns whether the triangle is clockwise or counter-clockwise.
	private final int determinant(	final int p1, final int p2, final int p3,
									final Point3[] pts) {
		double x1, x2, x3, y1, y2, y3;
		double determ;

		x1 = pts[p1].x;
		y1 = pts[p1].y;
		x2 = pts[p2].x;
		y2 = pts[p2].y;
		x3 = pts[p3].x;
		y3 = pts[p3].y;
		
		determ = (x2 - x1) * (y3 - y1) - (x3 - x1) * (y2 - y1);
		if (determ >= 0.0) return CCW;
		return CW;
	}
	
	// Answer 1 if no other point in the vertex list is inside
	// the triangle specified by the three points.  Returns
	// 0 otherwise.
	private final boolean noInterior(	final int p1, final int p2, final int p3,
										final Point3[] pts, final int[] vp, final int n,
										final int poly_or) {
		for (int i = 0; i < n; i++) {
			final int p = vp[i];			/* The point to test */
			if ((p == p1) || (p == p2) || (p == p3))
				continue;					/* Don't bother checking against yourself */
			if (   (determinant(p2, p1, p, pts) == poly_or)
					|| (determinant(p1, p3, p, pts) == poly_or)
					|| (determinant(p3, p2, p, pts) == poly_or) ) {
					continue;			/* This point is outside */
			} else {
					return false;		/* The point is inside */
			}
		}
		return true;					/* No points inside this triangle */
	}
}
