package com.exh.math.primitive;


/* Geometry operations.
 */
public class Geo {

	public final static double	DEGREES_TO_RADIANS = 0.017453;
	public final static double	RTOD = 57.2957795;				// radians to degrees
	
	public final static double	EPS = 0.00000000001;
	public final static double	SLOP = 0.00000000001;
	
	// Find the interesection of these lines using the line sweep algorithm.
	// http://www.topcoder.com/tc?module=Static&d1=tutorials&d2=geometry2
	// First convert form to A, B, C
	public final static void lineIntersection(	final Point3 a1, final Point3 a2,
												final Point3 b1, final Point3 b2,
												final Point3 out) {
		final double		A1 = a2.y - a1.y,
							B1 = a1.x - a2.x;
		final double		C1 = (A1*a1.x)+(B1*a1.y);
		final double		A2 = b2.y - b1.y,
							B2 = b1.x - b2.x;
		final double		C2 = (A2*b1.x)+(B2*b1.y);
		// Do the algorithm
		final double		det = (A1*B2) - (A2*B1);
		if (det == 0) {
			// Parallel
			out.x = a2.x;
			out.y = a2.y;
		} else {
			out.x = (B2*C1 - B1*C2)/det;
			out.y = (A1*C2 - A2*C1)/det;
		}
	}

	public final static void lineIntersection(	final double a1x, final double a1y,
												final double a2x, final double a2y,
												final double b1x, final double b1y,
												final double b2x, final double b2y,
												final Point3 out) {
		final double		A1 = a2y - a1y,
							B1 = a1x - a2x;
		final double		C1 = (A1*a1x)+(B1*a1y);
		final double		A2 = b2y - b1y,
							B2 = b1x - b2x;
		final double		C2 = (A2*b1x)+(B2*b1y);
		// Do the algorithm
		final double		det = (A1*B2) - (A2*B1);
		if (det == 0) {
			// Parallel
			out.x = a2x;
			out.y = a2y;
		} else {
			out.x = (B2*C1 - B1*C2)/det;
			out.y = (A1*C2 - A2*C1)/det;
		}
	}
	
	public final static void pointLineIntersection(	final double x, final double y,
													final double x0, final double y0,
													final double x1, final double y1,
													final Point3 out) {
		if (x1 - x0 == 0) {
			out.x = x0;
			out.y = y;
		} else if (y1 - y0 == 0) {
			out.x = x;
			out.y = y0;
		} else {
			final double		tg = -1 / ((y1 - y0) / (x1 - x0));
			out.x = (x1 * (x * tg - y + y0) + x0 * (x * - tg + y - y1)) / (tg * (x1 - x0) + y0 - y1);
			out.y = tg * out.x - tg * x + y;
		}
	}

	// From http://alienryderflex.com/intersect/
	public final static boolean lineSegmentIntersection(final double x1, final double y1,
														final double x2, final double y2,
														final double x3, final double y3,
														final double x4, final double y4,
														final Point3 out) { 
		final double	bx = x2 - x1,
						by = y2 - y1,
						dx = x4 - x3,
						dy = y4 - y3;
		final double	b_dot_d_perp = bx * dy - by * dx;
		if (b_dot_d_perp == 0) {
			return false;
		}
		final double	cx = x3 - x1,
						cy = y3 - y1;
		final double	t = (cx * dy - cy * dx) / b_dot_d_perp;
		if (t < 0 || t > 1) {
			return false;
		}
		final double	u = (cx * by - cy * bx) / b_dot_d_perp;
		if (u < 0 || u > 1) { 
			return false;
		}
		if (out != null) {
			out.x = x1+t*bx;
			out.y = y1+t*by;
		}
		return true;
	}

	// For cases where there are points that are allllmost in range
	public final static boolean sloppyLineSegmentIntersection(
										final double x1, final double y1,
										final double x2, final double y2,
										final double x3, final double y3,
										final double x4, final double y4,
										final Point3 out) { 
		final double	bx = x2 - x1,
						by = y2 - y1,
						dx = x4 - x3,
						dy = y4 - y3;
		final double	b_dot_d_perp = bx * dy - by * dx;
		if (b_dot_d_perp == 0) {
			return false;
		}
		final double	cx = x3 - x1,
						cy = y3 - y1;
		final double	t = (cx * dy - cy * dx) / b_dot_d_perp;
		if (t < 0 - SLOP || t > 1 + SLOP) {
			return false;
		}
		final double	u = (cx * by - cy * bx) / b_dot_d_perp;
		if (u < 0 - SLOP || u > 1 + SLOP) { 
			return false;
		}
		if (out != null) {
			out.x = x1+t*bx;
			out.y = y1+t*by;
		}
		return true;
	}

	public final static boolean lineSegmentIntersection(final Point3 a1, final Point3 a2,
														final Point3 b1, final Point3 b2,
														final Point3 out) {
		return lineSegmentIntersection(a1.x, a1.y, a2.x, a2.y, b1.x, b1.y, b2.x, b2.y, out);
	}

	// 0 = right, counter clockwise (90 = up)
	public final static double degree(double x2, double y2) {
		if (y2 == 0) {
			if (x2 > 0) return 0.0;
			else if (x2 < 0) return 180.0;
			else return 0;
		}
		if (x2 == 0) {
			if (y2 > 0) return 90.0;
			else if (y2 < 0) return 270.0;
		}
	
		final double		x1 = 0, y1 = 0;
		final double		m1 = 0;
		final double		m2 = slope(x1, y1, x2, y2);
		final double		tangent = (m2 - m1) / (1 + (m1 * m2));
		final double		radians = Math.atan(tangent);
		final double		degree = radians * RTOD;
	
		if (x2 < 0 && y2 > 0) return degree + 180;
		else if (x2 < 0 && y2 < 0) return degree + 180;
		else if (x2 > 0 && y2 < 0) return degree + 360;
		
		return degree;
	}
	
	// Given a line in point-slope form, find a new point on
	// that line that's dist away.
	public final static void newPtOnLine(	final double x, final double y, final double  m, final double dist,
											final Point3 newPt) {
		if (m == 0) {
			newPt.x = (float)(x + dist);
			newPt.y = (float)y;
			return;
		} else if (Double.isInfinite(m)) {
			newPt.x = (float)(x);
			newPt.y = (float)(y + dist);
			return;
		}

		/* This is basically a reverse engineering of the distance formula
		 * 		dist = sqrt(SQR(x2 - x1) + SQR(y2 - 1))
		 * It works by taking the final result -- the requested new distance,
		 * and deconstructing it into the xPart (SQR(x2 - x1) and the yPart,
		 * the sqrt of which can be added to the x and y values.
		 *
		 * The only trick is finding the xPart, which is just solved via
		 * the equation x + (x + (slope * slope)) = newDistance.
		 */
		final double		xPart = (sqr(dist) * (1 / (m * m))) / ( (1 / (m * m)) + 1);
		final double		yPart = xPart * (m * m);

		// Hmmm... just guessing...
		if (m < 0) {
			if (dist < 0) {
				newPt.x = (float)(x + Math.sqrt(xPart));
				newPt.y = (float)(y - Math.sqrt(yPart));
			} else {
				newPt.x = (float)(x - Math.sqrt(xPart));
				newPt.y = (float)(y + Math.sqrt(yPart));
			}
		} else {
			if (dist < 0) {
				newPt.x = (float)(x - Math.sqrt(xPart));
				newPt.y = (float)(y - Math.sqrt(yPart));
			} else {
				newPt.x = (float)(x + Math.sqrt(xPart));
				newPt.y = (float)(y + Math.sqrt(yPart));
			}
		}
	}
	
	public final static double sqr(final double v) {
		return v*v;
	}

	public final static double dist(final double x0, final double y0,
									final double x1, final double y1) {
		return Math.sqrt(((x0-x1)*(x0-x1)) + ((y0-y1)*(y0-y1)));
	}

	// Ignore that slow square root when we're only comparing
	// relative distances.
	public final static double distSquared(	final double x0, final double y0,
											final double x1, final double y1) {
		return ((x0-x1)*(x0-x1)) + ((y0-y1)*(y0-y1));
	}

	public final static double slope(final double x0, final double y0, final double x1, final double y1) {
		return (y1 - y0) / (x1 - x0);
	}

	public final static double perpendicularSlope(final double x0, final double y0, final double x1, final double y1) {
		return -((x1-x0) / (y1-y0));
	}

	public final static double asPerpendicular(final double m) {
		return -1/m;
	}

	public final static boolean parallel(final double m0, final double m1) {
		if (Double.isInfinite(m0)) return Double.isInfinite(m1);
		return m0 == m1;
	}
	
	/* Given an x of 0 to 1, answer a value of 0 to 1
	 * that's smoothed towards the edges -- a standard S curve.
	 */
	public static double sCurve(final double x) {
		return (float)((x) * (x) * (3.0 - 2.0 * (x)));
	}
	
	/* Given an x of 0 to 1:
	 * f(0) = 0, f(0.5) = 1, f(1) = 0.  The center is smoothed,
	 * tapering off rapidly at the ends.
	 */ 
	public static double humpCurve(final double x) {
		return ((x <= 0 || x >= 1) ? 0 : (x <= 0.5) ? sCurve((x) * 2) : sCurve(1 - ((x) - 0.5) * 2));
	}

	// Set closed if the last point is same as first, otherwise closed = false.
	// http://local.wasp.uwa.edu.au/~pbourke/geometry/polyarea/
	public static final double polygonArea(	final Point3[] pt, final int origSize,
											final boolean closed) {
		double			ans = 0;
		final int		newSize = origSize + (closed ? 0 : 1);
		for (int k=0; k<newSize; k++) {
			if (k >= origSize) k = 0;
			
			final int	j = (k + 1) % newSize;
			ans += pt[k].x * pt[j].y;
			ans -= pt[k].y * pt[j].x;
		}
		ans /= 2;
		return (ans < 0 ? -ans : ans);
	}
	
	public static boolean polygonContains(	final Point3[] pt, final int size,
											final double x, final double y) {
		int				k, j;
		boolean			ans = false;
		for (k = 0, j = size-1; k < size; j = k++) {
			final Point3	pk = pt[k],
							pj = pt[j];
			if (( ((pk.y <= y) && (y < pj.y))
					|| ((pj.y <= y) && (y < pk.y)))
					&& (x < (pj.x - pk.x) * (y - pk.y) / (pj.y - pk.y) + pk.x))
				ans = !ans;
		}
		return ans;
	}

	// Given a polygon defined by pts and size, answer the barycenter.
	public static void getBarycenter(	final Point3[] pts, final int size,
										final Point3 bary) {
		bary.x = 0;
		bary.y = 0;
		bary.z = 0;
		for (int k=0; k<size; k++) {
			bary.x += pts[k].x;
			bary.y += pts[k].y;
			bary.z += pts[k].z;
		}
		if (size > 1) {
			final double	d = (double)size;
			bary.x /= d;
			bary.y /= d;
			bary.z /= d;
		}
	}
}