package com.exh.math.primitive;

import java.util.Formatter;


public final class Point3 {
	public double					x, y, z;
	
	// For when we want whole numbers
	private final static boolean	WHOLE = false;
	// For when we want higher accuracy in display values
	private final static boolean	PRECISE = true;
	
	// There's no reason to let coords get too precise, since
	// it all gets rendered into pixels anyway.
	private final static double		SLOP = 0.01;

	public final static Point3[]	newArray(final int size) {
		final Point3[]				a = new Point3[size];
		for (int k=0; k<size; k++) a[k] = new Point3();
		return a;
	}
	
	public Point3() {
	}

	public Point3(final Point3 o) {
		set(o);
	}
	
	public Point3(final double X, final double Y, final double Z) {
		set(X, Y, Z);
	}

	public void set(final Point3 o) {
		x = o.x;
		y = o.y;
		z = o.z;
	}

	public void set(final double X, final double Y, final double Z) {
		x = X;
		y = Y;
		z = Z;
	}
	
	public void set(final Point3 o, final double Z) {
		x = o.x;
		y = o.y;
		z = Z;
	}
	
	public boolean isNull() {
		return x == 0 && y == 0 && z == 0;
	}

	// Client equivalence test
	public boolean sameAs(final Point3 pt, final double eps) {
		if (this == pt || (x == pt.x && y == pt.y)) return true;
		return Math.abs(x-pt.x) < eps && Math.abs(y-pt.y) < eps;
	}

	// A weak equivalence test
	public boolean similarTo(final Point3 pt) {
		if (this == pt || (x == pt.x && y == pt.y)) return true;
		return Math.abs(x-pt.x) < SLOP && Math.abs(y-pt.y) < SLOP;
	}
	
	// A strong equivalence test
	public boolean sameAs(final Point3 pt) {
		if (this == pt || (x == pt.x && y == pt.y)) return true;
		return Math.abs(x-pt.x) < Geo.SLOP && Math.abs(y-pt.y) < Geo.SLOP;
	}
	
	public Point3 add(final Point3 b) {
		x += b.x;
		y += b.y;
		z += b.z;
		return this;
	}

	public void asAdd(final Point3 a, final Point3 b) {
		x = a.x+b.x;
		y = a.y+b.y;
		z = a.z+b.z;
	}

	public void asSub(final Point3 a, final Point3 b) {
		x = a.x-b.x;
		y = a.y-b.y;
		z = a.z-b.z;
	}
	
	public void asMid(final Point3 a, final Point3 b) {
		x = (a.x+b.x) / 2;
		y = (a.y+b.y) / 2;
		z = (a.z+b.z) / 2;
	}
	
	// Obsolete
	public void add(final Point3 a, final Point3 b) {
		asAdd(a, b);
	}

	// Obsolete
	public void sub(final Point3 a, final Point3 b) {
		asSub(a, b);
	}
	
	public void mult(final double v) {
		x *= v;
		y *= v;
		z *= v;
	}
	
	public Point3 div(final double v) {
		x /= v;
		y /= v;
		z /= v;
		return this;
	}
	
	public void normalize() {
		double	nrml = Math.sqrt((x*x) + (y*y) + (z*z));
		if (nrml == 0) return;
		x = x / nrml;
		y = y / nrml;
		z = z / nrml;
	}
	
	public double length() {
		return Math.sqrt(x*x + y*y + z*z);
	}
	
	public double distSquared(final Point3 b) {
		return Geo.distSquared(x, y, b.x, b.y);
	}
	
	// Distance in 3d
	public double dist3(final Point3 b) {
		return Math.sqrt( ((x-b.x)*(x-b.x)) + ((y-b.y)*(y-b.y)) + ((z-b.z)*(z-b.z)) );
	}

	public double dist3Squared(final Point3 b) {
		return ((x-b.x)*(x-b.x)) + ((y-b.y)*(y-b.y)) + ((z-b.z)*(z-b.z));
	}

	public final double dot2d(final Point3 v) {
		return x * v.x + y * v.y;
	}
	
	public final double dot3(final Point3 v) {
		return x * v.x + y * v.y + z * v.z;
	}
	
	public void asVector(final Point3 a, final Point3 b) {
		x = b.x-a.x;
		y = b.y-a.y;
		z = b.z-a.z;
	}
		
	public void asCrossProduct(final Point3 a, final Point3 b) {
		x = a.y*b.z - b.y*a.z;
		y = a.z*b.x - b.z*a.x;
		z = a.x*b.y - b.x*a.y;
	}

	// Given two points, answer a value between them based
	// on pos.  A pos of 0 is a, 0.5 is the mid point, 1 is b.
	public void blend(final Point3 a, final Point3 b, final double pos) {
		x = (a.x*(1-pos)) + (b.x*pos);
		y = (a.y*(1-pos)) + (b.y*pos);
		z = (a.z*(1-pos)) + (b.z*pos);
	}

	public String toString() {
		final Formatter		fmt = new Formatter();
		if (WHOLE) return fmt.format("(%d, %d, %d)", (int)x, (int)y, (int)z).toString();
		if (PRECISE) return fmt.format("(%.5f, %.5f, %.5f)", x, y, z).toString();
		return fmt.format("(%.2f, %.2f, %.2f)", x, y, z).toString();
	}
	
	/* ORDER
	 * Utility for altering the order of point arrays.
	 ************************************************/
	public final static class Order {
		private double[]		mD = null;
	
		// Check if the points are ordered clockwise by
		// looking at the degree of each in relation to
		// the barycenter.
		public boolean isCw(final Point3[] pts, final int size) {
			if (!verify(pts)) return false;
			cacheDegrees(pts, size);

			// Find the lowest
			double			lowD = mD[0];
			int				lowI = 0;
			for (int k=1; k<size; k++) {
				if (mD[k] < lowD) {
					lowD = mD[k];
					lowI = k;
				}
			}
			// Make sure they're in order
			for (int step = 1; step < size; step++) {
				int			nextI = lowI + 1;
				if (nextI >= size) nextI = 0;
				if (mD[nextI] < lowD) return false;
				lowD = mD[nextI];
				lowI = nextI;
			}
			return true;
		}

		public void makeCcw(final Point3[] pts) {
			if (pts == null) return;
			makeCcw(pts, pts.length);
		}

		public void makeCcw(final Point3[] pts, final int size) {
			if (!verify(pts, size)) return;
			cacheDegrees(pts, size);
			
			// Sort the degrees, highest to lowest
			Point3		pt;
			double		d;
			boolean		more = true;
			while (more) {
				more = false;
				for (int k = 1; k < size; k++) {
					if (mD[k] > mD[k-1]) {
						d = mD[k-1];	mD[k-1] = mD[k];	mD[k] = d;
						pt = pts[k-1];	pts[k-1] = pts[k];	pts[k] = pt;
						more = true;
					}
				}
			}
		}
		
		public void makeReverse(final Point3[] pts) {
			int					dst = pts.length-1;
			for (int src=0; src<dst; src++) {
				// swap
				final Point3	pt = pts[src];
				pts[src] = pts[dst];
				pts[dst] = pt;

				dst--;
			}
		}
		
		private final boolean verify(final Point3[] pts) {
			if (pts == null) return false;
			return verify(pts, pts.length);
		}
		
		private final boolean verify(final Point3[] pts, final int size) {
			if (pts == null) return false;
			if (mD == null || mD.length < size) {
				try {
					mD = new double[size];
				} catch (Exception ex) {
					return false;
				}
			}
			return true;
		}
		
		private final void cacheDegrees(final Point3[] pts, final int size) {
			// Find the barycenter
			double		baryX = 0,
						baryY = 0;
			for (int k=0; k<size; k++) {
				baryX += pts[k].x;
				baryY += pts[k].y;
			}
			baryX /= (double)size;
			baryY /= (double)size;
			// Find each point's degree
			for (int k=0; k<size; k++) {
				mD[k] = Geo.degree(pts[k].x - baryX, pts[k].y - baryY);
			}
		}
	}
}
