package com.exh.math.tests;

import com.exh.math.primitive.Geo;
import com.exh.math.primitive.Point3;
import com.exh.math.primitive.Polygon;

/* Given polygons a and b, answer true if a is external
 * to b.  This test is used when we know the polygons
 * overlap, but not by how much, or to weed out polygons
 * that are merely adjacent, sharing a colinear line,
 * from ones that actually infringe on each other.
 * 
 * NOTE:  This algorithm is basically ad hoc and tuned
 * to my needs.  The ideal solution would be to union
 * the polygons, find the area, and compare that with
 * the area of the individual polys, but I'm assuming
 * this is faster and hoping it's accurate for every
 * condition encountered by the framework.
 */
public final class PolyIsExternal {

	private final SegSeg			mSS = new SegSeg();
	private final LineIsColinear	mLiC = new LineIsColinear();
	
	private final Point3			mPt = new Point3();
	
	private final static double		EPS = 0.1;
	
	private final static double		INTRUDES_D = 0.5;
	
	private final static double		TOLERANCE = 0.5;
	private final static double		TOLERANCE_SQUARED = TOLERANCE*TOLERANCE;

	public PolyIsExternal() {
		mSS.mEps = EPS;
		mLiC.mEps = EPS;
	}

	public boolean on(final Polygon a, final Polygon b) {
/*
System.out.println("POLY IS EXTERNAL");
System.out.println("\ta"); a.print();
System.out.println("\tb"); b.print();
System.out.println("external(a,b)=" + isExternalTo(a, b) + " external(b,a)=" + isExternalTo(b,a));
*/
		return isExternalTo(a, b) && isExternalTo(b, a);
	}

	private boolean isExternalTo(final Polygon a, final Polygon b) {
		// Weed out a specific case -- b's barycenter is inside a.
		b.getBarycenter(mPt);
		if (Geo.polygonContains(a.mPt, a.mSize, mPt.x, mPt.y)) return false;
		
		Point3					prvA = a.mPt[a.mSize-1];
		for (int k=0; k<a.mSize; k++) {
			final Point3		nxtA = a.mPt[k];
			Point3				prvB = b.mPt[b.mSize-1];
			for (int j=0; j<b.mSize; j++) {
				final Point3	nxtB = b.mPt[j];
				// The base line for having an interior test is that the
				// lines must intersect, and not be colinear.
				if (mSS.on(prvA, nxtA, prvB, nxtB) && !mLiC.on(prvA, nxtA, prvB, nxtB)) {
//		System.out.println("intersect " + prvA + " " + nxtA + " to " + prvB + " " + nxtB);
					// If intersecting with an endpoint, I can skip.
					if (!isEndpoint(mSS.mHit, prvB, nxtB)) {
//		System.out.println("\tintrudes=" + intrudes(prvA, nxtA, mSS.mHit, b));
						if (intrudes(prvA, nxtA, mSS.mHit, b)) return false;
					}
				}
				prvB = nxtB;
			}			
			prvA = nxtA;
		}
		return true;
	}

	private boolean isEndpoint(	final Point3 intersection,
								final Point3 a0, final Point3 a1) {
		return intersection.dist3Squared(a0) < TOLERANCE_SQUARED
				|| intersection.dist3Squared(a1) < TOLERANCE_SQUARED;
	}

	// A line intrudes on the polygon if there is a point on either
	// side of the intersection that is interior.
	private boolean intrudes(	final Point3 a0, final Point3 a1,
								final Point3 intersection,
								final Polygon poly) {
		double					d;
		// If a0 is sufficiently far from the intersection to do a test...
		if ((d = a0.dist3(intersection)) > INTRUDES_D) {
			mPt.blend(a0, intersection, (d-INTRUDES_D) / d);
//	System.out.println("a0 d=" + d + " pt=" + mPt);
			if (Geo.polygonContains(poly.mPt, poly.mSize, mPt.x, mPt.y)) return true;
		}
		// If a1 is sufficiently far from the intersection to do a test...
		if ((d = a1.dist3(intersection)) > INTRUDES_D) {
			mPt.blend(a1, intersection, (d-INTRUDES_D) / d);
//	System.out.println("a1 d=" + d + " pt=" + mPt);
			if (Geo.polygonContains(poly.mPt, poly.mSize, mPt.x, mPt.y)) return true;
		}
		return false;
	}
}
