package com.exh.math.tests;

import java.util.Comparator;
import java.util.TreeSet;

import com.exh.math.primitive.Point3;
import com.exh.math.primitive.Polygon;

/* PT-IN-POLY
 * Determine if the point lies in the polygon, in 3D.  It's
 * assumed that the point lies on the same plane as the poly.
 * Do this by finding a line from the pt to the poly, then
 * extending that line beyond the bounds, finding all
 * intersections, and determining what lines enclose the point.
 * NOTE:  Pts on an edge are considered inside.
 * NOTE:  If the polygon is a point or line, this test will
 * always fail, even though the nice thing to do would be to
 * check if it falls on the point or line.
 */
public final class PtInPoly {

public boolean mPrint = false;

	private final Point3				mA0 = new Point3(),
										mA1 = new Point3(),
										mDelta0 = new Point3(),
										mDelta1 = new Point3();
	
	private final SegSeg				mSS = new SegSeg();
	
	// Handle sorting all the intersections, and placing
	// the original point in context
	private final PointComparator		mPtCmp = new PointComparator();
	private final TreeSet<Intersection>	mSort = new TreeSet<Intersection>(mPtCmp);

	// PT-IN-POLY
	// --------------------------------------------
	public boolean on(final Point3 pt, final Polygon poly) {
		if (poly.mSize < 3) return false;
		constructSegment(pt, poly);
if (mPrint) System.out.println("\tseg " + mA0 + ", " + mA1);
mSS.mLL.mPrint = mPrint;
		mPtCmp.mOrigin = mA0;
		mSort.clear();
		mSort.add(newSrc(pt));
		// And all the intersection
		Point3				prv = poly.mPt[poly.mSize-1];
		// Take the largest distance, so we have the most "room"
		// and are less influenced by 
		for (int k = 0; k<poly.mSize; k++) {
			final Point3	nxt = poly.mPt[k];
if (mPrint) System.out.println("\tSS on prv " + prv + " nxt " + nxt);
			if (mSS.on2(mA0, mA1, prv, nxt)) {
				mSort.add(newHit(mSS.mHit));
			}
	if (mPrint) System.out.println("\t d=" + mSS.mDistanceToHit);
			prv = nxt;
		}
for (Intersection i : mSort) {
	if (mPrint) System.out.println("\t" + i);
}
		// Now apply the rules -- if the pt is on an edge, obviously
		// it's not an interior
		if (mSort.first().mSrc != null || mSort.last().mSrc != null) return false;
		// Otherwise count the number of hits surrounding the source, odd
		// for interior, even for exterior
		boolean				exterior = true;
		for (Intersection i : mSort) {
			exterior = !exterior;
			if (i.mSrc != null) return exterior;
		}
		// ERROR CONDITION, actually.
		return false;
	}
	
	// Construct a line that will intersect the poly somewhere
	// and extend through it fully.
	private void constructSegment(final Point3 pt, final Polygon poly) {
		mA0.set(pt);
		mA1.set(furthest(pt, poly));
		mDelta0.asSub(mA0, mA1);
		mDelta1.asSub(mA1, mA0);
		// Expand the deltas so we're guaranteed to encompass the poly.
		mDelta0.mult(4);
		mDelta1.mult(4);
		// Extend the line according to the delta
		mA0.asAdd(mA0, mDelta0);
		mA1.asAdd(mA1, mDelta1);
	}
	
	private Point3 furthest(final Point3 pt, final Polygon poly) {
		Point3				ans = poly.mPt[0];
		double				d = pt.dist3Squared(ans);
		for (int k=1; k<poly.mSize; k++) {
			final double	nextD = pt.dist3Squared(poly.mPt[k]);
			if (nextD > d) {
				ans = poly.mPt[k];
				d = nextD;
			}
		}
		return ans;
	}
	
	private Intersection newSrc(final Point3 src) {
		final Intersection	i = new Intersection();
		i.mSrc = src;
		return i;
	}
	
	private Intersection newHit(final Point3 hit) {
		final Intersection	i = new Intersection();
		i.mHit.set(hit);
		return i;
	}
	
	// INTERSECTION
	// --------------------------------------------
	private final static class Intersection {
		Point3				mSrc = null;
		final Point3		mHit = new Point3();
		
		Point3 getPt() {
			if (mSrc != null) return mSrc;
			return mHit;
		}
		
		public String toString() {
			if (mSrc != null) return "src=" + mSrc;
			return "hit=" + mHit;
		}
	}

	// POINT-COMPARATOR
	// For comparing point distance from an origin
	// --------------------------------------------
	private final static class PointComparator implements Comparator<Intersection> {
		
		Point3				mOrigin;
		
 		public int compare(Intersection o1, Intersection o2) {
 			final double	d1 = mOrigin.dist3Squared(o1.getPt()),
 							d2 = mOrigin.dist3Squared(o2.getPt());
 			if (d1 < d2) return -1;
 			if (d1 == d2) return 0;
 			return 1;
 		}
 	}

}
