package com.exh.math.tests;

import com.exh.math.primitive.Geo;
import com.exh.math.primitive.Point3;
import com.exh.math.primitive.Segment;

/* Class to determine if and where a ray interesects a tri.
 * Adapted from http://www.softsurfer.com/Archive/algorithm_0105/algorithm_0105.htm
 */
public final class RayIntersectsTri {

	// When the function is done, this will answer whether or
	// not the actual segment intersected (provided the test succeeded).
	public boolean					mSegmentIntersects = false;

	public final static int			ERROR = -1;
	public final static int			NO_INTERSECT = 0;
	public final static int			INSIDE = 1;
	// The following are actual success conditions, and if
	// they're true, then out is valid
	public final static int			INTERSECT_PLANE = 2;
	public final static int			INTERSECT_TRI = 3;
	
	private final Point3			mU = new Point3(),		// Triangle vectors
									mV = new Point3(),
									mN = new Point3();
	private final Point3			mDir = new Point3(),	// Ray vectors
									mW0 = new Point3(),
									mW = new Point3();
	
	private final static double		EPS = 0.0000000001;
	
	/* Answer a value determining the type of intersection.  Out
	 * is valid on an answer of INTERSECT_PLANE or INTERSECT_TRI.
	 * Tri is CCW.
	 */
	public int on(	final Segment ray, final Point3 ptA, final Point3 ptB,
					final Point3 ptC, final Point3 out) {
		// get triangle edge vectors and plane normal
		mU.sub(ptB, ptA);
		mV.sub(ptC, ptA);
		mN.asCrossProduct(mU, mV);
		// Don't deal with degenerate triangles
		if (mN.isNull()) return ERROR;

		mDir.sub(ray.mPt[1], ray.mPt[0]);
		mW0.sub(ray.mPt[0], ptA);
		final double		a = -mN.dot3(mW0);
		final double		b = mN.dot3(mDir);
		if (Math.abs(b) < EPS) {			// ray is parallel to triangle plane
			if (a == 0) return INSIDE;		// ray lies in triangle plane
			else return NO_INTERSECT;		// ray disjoint from plane
		}

		// get intersect point of ray with triangle plane
		final double		r = a / b;
		if (r < 0.0) return NO_INTERSECT;	// ray goes away from triangle
		// for a segment, also test if (r > 1.0) => no intersect
		mSegmentIntersects = (r < 1.0 + EPS);
		
		// Intersect point of ray and plane
		out.set(ray.mPt[0].x + r * mDir.x,
				ray.mPt[0].y + r * mDir.y,
				ray.mPt[0].z + r * mDir.z);

		// is the intersection inside the triangle?
		final double		uu = mU.dot3(mU),
							uv = mU.dot3(mV),
							vv = mV.dot3(mV);
		mW.sub(out, ptA);
		final double		wu = mW.dot3(mU),
							wv = mW.dot3(mV);
		final double		D = uv * uv - uu * vv;

		// get and test parametric coords
		final double		s = (uv * wv - vv * wu) / D;
		if (s < -Geo.EPS || s > 1.0 + Geo.EPS) return INTERSECT_PLANE;			// out is outside T
		final double		t = (uv * wu - uu * wv) / D;
		if (t < -Geo.EPS || (s + t) > 1.0 + Geo.EPS) return INTERSECT_PLANE;	// out is outside T
		
		return INTERSECT_TRI;
	}
	
	// Only use after performing the setup on a tri on the
	// same plane as this tri.
	int onTri(	final Point3 a, final Point3 b, final Point3 c,
				final Point3 out) {
		mU.sub(b, a);
		mV.sub(c, a);
		
		final double		uu = mU.dot3(mU),
							uv = mU.dot3(mV),
							vv = mV.dot3(mV);
		mW.sub(out, a);
		final double		wu = mW.dot3(mU),
							wv = mW.dot3(mV);
		final double		D = uv * uv - uu * vv;

		// get and test parametric coords
		final double		s = (uv * wv - vv * wu) / D;
		if (s < -Geo.EPS || s > 1.0 + Geo.EPS) return INTERSECT_PLANE;			// out is outside T
		final double		t = (uv * wu - uu * wv) / D;
		if (t < -Geo.EPS || (s + t) > 1.0 + Geo.EPS) return INTERSECT_PLANE;	// out is outside T
		
		return INTERSECT_TRI;
	}

}
